// ======================================
//  NAME    : fwTell1 Library
//
//  VERSION : 2
//
//  REVISION: 0
//
//  RELEASE : 2007/Dec/05
//
//  PLATFORM: PVSS II
//
//  AUTHOR  : Stefan Koestner
//
//  EMAIL   : stefan.koestner@cern.ch
// ======================================


#uses "fwCcpc.ctl"
#uses "fwHw.ctl"
#uses "fwTell1.ctl"



/*! \mainpage fwTell1 Library Documentation
 *
 *  Library of PVSS functions frequently used for operating the TELL1-boards
 *
 *  Modifications: none
 *
 *  Usage: JCOP framework internal, public
 *
 *  PVSS manager usage: VISION, CTRL
 *
 *  \n
 *  \n
 *  \li \ref SectionConstants
 *  \n
 *  \n
 *  \li \ref SectionArbitrary
 *  \n
 *  \n
 *  \li \ref SectionHardware
 *  \n
 *  \n
 *  \li \ref SectionCCPC
 *  \n
 *  \n
 *  \li \ref SectionCheck
*/






// ======================================
//  CONSTANTs & GLOBAL VARs
// ======================================

/** @defgroup SectionConstants Constants
 *  List of constants and global variables used in the framework component.
 *  @{
*/




/** This global variable represents the board type.
 *  \n
*/
global string				FWTELL1_BOARDTYPE = "NOT_DEF";





/** @} */


//const string			FWCCPC_CONFIG_NAME = "CcpcConfig";
//const string			FWCCPC_PREFIX = "ccpc/";
//const int         FWCCPC_SERVER_VERSION = 220;
//const int					FWCCPC_TIME_OUT = 10;
//global dyn_string fwCcpcList;	
//global int				fwCcpcStateManager;
//global bool				fwCcpcChangeScriptsFlag;







// ======================================
//  ARBITRARY FUNCTIONs
// ======================================

/** @defgroup SectionArbitrary Arbitrary Functions
 *  List of frequently used functions in the fwTell1 framework component.
 *  @{
*/




/** This function clones many Tell1 boards from one original and assigns the credit card PC.  
 * 
 *  \param [in] Original Tell1 board from which to clone. 
 *  \param [in] Clones The names of the new Tell1 boards. 
 *  \param [in] CCPCs The credit card PCs to be assigned to Tell1 boards. 
 *  \return \e int the errorcode of this library: 0 if no error occured 
 *                                               -1 ECS error
 *                                               >0 number of boards from where the ccpc name could not be read back after cloning
 * 
 *  \b Example: 
 * 
 *  \code 
 *  (...) 
 *  if (fwTell1_iniGlueCard("pclbcecs03") == 0)  
 *     (...) 
 *  else 
 *     DebugTN("Error found!"); 
 *  (...) 
 *  \endcode 
 * 
 *  \n 
 *  \n 
*/ 


int fwTell1_cloneManyTell1Boards(string Original, dyn_string Clones, dyn_string CCPCs){
  int i;
  //string Original = "myTell1_ST";
  //dyn_string Clones = makeDynString("TT_1", "TT_2", "TT_3");
  //dyn_string CCPCs = makeDynString("tell1", "tell2", "tell3");
  dyn_anytype dummy;
  dynClear(dummy);
  int err;
  
  for(i = 1; i <= dynlen(Clones); i++){
     dynAppend(dummy, CCPCs[i]);
     //DebugN(dummy);
     DebugN(Clones[i] + " has ccpc name: " + CCPCs[i]);
     if(!fwHw_clone(Original, Clones[i]))
       return -1;
     if(!fwHw_setCommonSettings(Clones[i],dummy,true))
       return -1;
     dynClear(dummy); 
  }
  
  DebugN("Boards are created!");
  
   for(i = 1; i <= dynlen(Clones); i++){
     if(!fwHw_getCommonSettings(Clones[i],dummy))
       return -1;
     //DebugN(dummy);
     if(dummy[1] != CCPCs[i])
       err++;
     dynClear(dummy);
   }
  
   if(err)
     DebugN("Error ocurred when reading back CCPC name from common settings!");
   else
     DebugN("CCPC names could be read back without error!");
   return err;
 }
 




/** This function assembles various sub-recipes into one single recipe. 
 *  It takes care of masked operations: meaning that in the recipes different bits can be set in one register
 *  without overwriting each other. 
 * 
 *  \param [in] hardware The hardware name to which all recipes are related to. 
 *  \param [in] recipes The names of all sub-recipes which shall be assembled. 
 *  \param [in] RecipeName The name of the final, combined recipe. 
 *  \param [in] dpType The datapoint type corresponding to the hardware.
 *  \param [in] RecipeType The recipe type of the new, combined recipe.
 *  \return \e int the errorcode of this library: 0 if no error occured 
 *                                               -1 ECS error
 * 
 *  \b Example: 
 * 
 *  \code 
 *  (...) 
 *  if (fwTell1_iniGlueCard("pclbcecs03") == 0)  
 *     (...) 
 *  else 
 *     DebugTN("Error found!"); 
 *  (...) 
 *  \endcode 
 * 
 *  \n 
 *  \n 
*/ 
int fwTell1_AssembleAndCreateRecipes(string hardware, dyn_string recipes, string RecipeName, string dpType, string RecipeType){ 
   
 
  dyn_dyn_string Registers; 
  dyn_dyn_int Registers_Size; 
  dyn_dyn_int Registers_Type; 
  dyn_string  HardwRegisters_ALL, regs;  
  dyn_int   Types_ALL;
  dyn_dyn_anytype RecipeData1, RecipeData2,RecipeDataALL; 
  dyn_dyn_anytype data1, mask1; 
  dyn_dyn_anytype data2, mask2; 
  dyn_dyn_anytype dummy; 
  int j,i; 
  int num; 
  bool SameRegistersFlag; 
  dyn_string dum; 
  
  DebugTN("=========================================================================="); 
  DebugTN("====================ASSEMBLE=THE=DIFFERENT=RECIPES========================");  
 
   DebugN(recipes);
  if(dynlen(recipes) > 1){     
	if(!fwHw_getRecipe(hardware, recipes[1], Registers[1], RecipeData1))
          return -1;  
        DebugN(Registers[1]); 
 	for(num = 1; num <= dynlen(recipes) - 1 ; num++){ 
	   	if(dynlen(Registers[num])>0){ 
	     		if(fwTell1_getRegInfo(hardware, Registers[num], Registers_Size[num], Registers_Type[num]) == -1)
                          return -1; 
                        DebugN(Registers_Type);
	                if(!fwHw_getRecipe(hardware, recipes[num+1], Registers[num+1], RecipeData2))
                          return -1; 
	                if(dynlen(Registers[num])>0){ 
 
				//fwTell1_getRegInfo  givs the size and the type of each registers 
				if(fwTell1_getRegInfo(hardware, Registers[num+1], Registers_Size[num+1], Registers_Type[num+1]) == -1)
                                  return -1;
				 
	   			for(i = 1; i <= dynlen(Registers[num]); i++){ 
                                        dynClear(dummy); 
	   				SameRegistersFlag = false; 
	     				for(j = 1; j<=  dynlen(Registers[num+1]); j++){	      					  
	                                        if(Registers[num][i] == Registers[num+1][j]){ 							 
							SameRegistersFlag = true; 
							break ; 
						}               
					} 
					if(SameRegistersFlag){ 
 
						dynClear(dummy); 
						dynClear(data1); 
	    					dynClear(mask1); 
						dynClear(data2); 
						dynClear(mask2); 
 
						//fwTell1_updateVar gives back data and mask 
	    					fwTell1_updateVar(RecipeData1[i], Registers_Size[num][i], data1[i], mask1[i]); 
					        fwTell1_updateVar(RecipeData2[j], Registers_Size[num+1][j], data2[j], mask2[j]);    
    
						DebugTN("re-create recipes for register: "+Registers[num][i]); 
 
						//update of the data and the mask (according to the two old ones) 
					        fwTell1_writeToVar(data1[i],mask1[i],data2[j],mask2[j]) ; 
 
						dynAppend(dummy[i], data1[i]); 
						dynAppend(dummy[i], mask1[i]);    
 
						//remove in the other dyn_string num+1 
					        dynRemove(Registers[num+1],j); 
 					        dynRemove(Registers_Type[num+1],j); 
					        dynRemove(RecipeData2,j); 
 
 
					}else{ 
						dynAppend(dummy[i], RecipeData1[i]); 
					} 
   
					dynAppend(HardwRegisters_ALL, Registers[num][i]); 
					dynAppend(Types_ALL,Registers_Type[num][i]); 
					dynAppend(RecipeDataALL, dummy[i]); 
 
					//for the Recipes Type ".SL.blabla" 
					dynAppend(Registers[dynlen(recipes)+1], substr(Registers[num][i],strlen(hardware)));     
			    		 
				} 
				for(i = 1; i <= dynlen(Registers[num+1]); i++){       
					dynAppend(HardwRegisters_ALL, Registers[num+1][i]); 
					dynAppend(Types_ALL,Registers_Type[num+1][i]); 
					dynAppend(RecipeDataALL, RecipeData2[i]); 
					dynAppend(Registers[dynlen(recipes)+1], substr(Registers[num+1][i],strlen(hardware))); 
				} 
				 
			} 
			dynClear(Registers[num+1]); 
			dynClear(Registers_Size[num+1]); 
			dynClear(Registers_Type[num+1]); 
			dynClear(RecipeData2);  
			Registers[num+1] = HardwRegisters_ALL; 
			RecipeData1 = RecipeDataALL; 
                        //warning
			if(num < dynlen(recipes) - 1 ){ 
				dynClear(HardwRegisters_ALL); 
				dynClear(Types_ALL); 
				dynClear(RecipeDataALL); 
				dynClear(Registers[dynlen(recipes)+1]); 
			} 
 
		}else{ 
 
 		DebugTN("No register in this recipe: "+recipes[num]+" ?!?!"); 
 
		} 
 
	} 
 
   }else{ 
 
   DebugTN("Less than two recipes were given."); 
 
   } 
 
   DebugTN(HardwRegisters_ALL);
  
   
   //this is a dirty bugfix - the problem arrises elsewhere!!!
   for(i=1; i<=dynlen(HardwRegisters_ALL); i++){
    dum = strsplit(HardwRegisters_ALL[i],":");
            if(dynlen(dum) >= 2)
              HardwRegisters_ALL[i] = dum[2];
      HardwRegisters_ALL[i] = substr(HardwRegisters_ALL[i],strlen(hardware));
          }
   
          
          
     //DebugTN(Registers[dynlen(recipes)+1]);  
     //DebugTN("new");      
            
   //fwHw_createRecipeType(RecipeType,dpType1,Registers[dynlen(recipes)+1]);
    if(!fwHw_createRecipeType(RecipeType,dpType,HardwRegisters_ALL))
      return -1;
   DebugN("Registers belonging to recipe type: " + RecipeType);  

   if(!fwHw_getRecipeTypeInfo(RecipeType, dpType, regs))
     return -1; 
   DebugN(regs);    
  
   if(!fwHw_saveRecipe(hardware, RecipeType, RecipeName, HardwRegisters_ALL, Types_ALL, RecipeDataALL))
     return -1;
   
  // DebugN(hardware, RecipeType, RecipeName); 
  // DebugN(hardware, RecipeType, RecipeName); 
   DebugTN("=========================END=OF=THE=ASSEMBLY=============================="); 
   DebugTN("=========================================================================="); 
 
   return 0;
 }







/** This function cuts out a bit-array from a hexadecimal string.
 *
 *  \param [in] pHex hexadecimal number (represented as a string) from which an array of bits is cut out
 *  \param [in] pUpper the most significant bit (counted from zero - conform with Guido's ECS docu!)
 *  \param [in] pLower the least significant bit (counted from zero - conform with Guido's ECS docu!)
 *  \return \e string the bit array represented as a hexadecimal number (string not integer!)
 *
 *  \b Example: 
 *
 *  \code
 *  (...)
 * // read a register and represent a few adjacent bits out of the full register as a number on the panel
 * dyn_dyn_char data;  
 * dyn_int status;
 * dyn_string registersDP;
 * dynClear(registersDP);
 * dynAppend(registersDP, "dist_1:myTell1_OT.SL.CMNCTRL.SLCtrlReg1.readings"); 
 * fwCcpc_read(registersDP, data, status)
 *  SLCtrlReg1.text = fwTell1_BitsFromString(fwCcpc_convertByteToHex(data[2]),31,24);
 *  (...)
 *  \endcode
 *
 *  \n
 *  \n
*/

string fwTell1_BitsFromString(string pHex, int pUpper, int pLower){
   string casted;
   int number;
   int dummy;
   int mask = 0;
   int i;
 
   number = fwCcpc_convertHexToDec(pHex);
   number = number >> pLower;   
   
   for(i = 0; i <= (pUpper - pLower); i++){
      dummy = 1 << i;
      mask = dummy + mask;
   }
   dummy = number & mask;
   casted = fwCcpc_convertDecToHex(dummy);
   return casted;
}



/** This function cuts out a bit-array from an integer.
 *
 *  \param [in] pNumber an integer number from which an array of bits is cut out
 *  \param [in] pUpper the most significant bit (counted from zero - conform with Guido's ECS docu!)
 *  \param [in] pLower the least significant bit (counted from zero - conform with Guido's ECS docu!)
 *  \return \e int the bit array represented as an integer number 
 *
 *  \b Example:
 *
 *  \code
 *  (...)
 * // read a register and represent a few adjacent bits out of the full register as a number on the panel
 * dyn_dyn_char data;  
 * dyn_int status;
 * dyn_string registersDP;
 * dynClear(registersDP);
 * dynAppend(registersDP, "dist_1:myTell1_OT.SL.CMNCTRL.SLCtrlReg1.readings"); 
 * fwCcpc_read(registersDP, data, status)
 *  SLCtrlReg1.text = fwTell1_BitsFromString(fwCcpc_convertByteToDec(data[2]),31,24);
 *  (...)
 *  \endcode
 *
 *  \n
 *  \n
*/

int fwTell1_BitsFromInt(int pNumber, int pUpper, int pLower){
   int dummy;
   int mask = 0;
   int i;
   pNumber = pNumber >> pLower;   
   
   for(i = 0; i <= (pUpper - pLower); i++){
      dummy = 1 << i;
      mask = dummy + mask;
   }
   dummy = pNumber & mask;
   return dummy;  
}





/** @} */






// ======================================
//  COMMON HARDWARE FUNCTIONs
// ======================================

/** @defgroup SectionHardware Hardware Functions
 *  The parameters of these functions will always be the hardware name (datapoint name).
 *  The functions will just work, if the boards are subscribed!
 *  @{
*/



/** TELL1 boards of different subdetectors can be recognised by reading various bits 
 *  of a special register. The global variable FWTELL1_BOARDTYPE can be used to store the
 *  returned string which is a short acronym for the board type. By default this variable is "NOT_DEF".
 *
 *  \param [in] Tell1Name the name of the hardware (Tell1 board).
 *  \param [out] TypeOfBoard returns the abriations of the board types.
 *  \return \e int the errorcode of this library: 0 if no error occured
 *                                               -1 register could not be read
 *
 *
 *  \b Example:
 *  
 *  \code
 *  (...)
 *  if (FWTELL1_BOARDTYPE == "NOT_DEF")
 *     if(fwTell1_TypeOfBoard("myTell1", FWTELL1_BOARDTYPE) == -1)
 *        DebugN("Reading of register not successfull! Board not identified!");
 *  (...)
 *  \endcode
 *
 *  \n
 *  \n
*/

int fwTell1_TypeOfBoard(string Tell1Name, string &TypeOfBoard){
   
   string dummy;
   int address;
   int status;
   dyn_dyn_char data;  
   dyn_int statusG;
   dyn_string registersDP;
   
   //address = fwCcpc_convertHexToDec("4001000");
   //status = fwCcpc_LBUSRead(CCPC, address, FWCCPC_LBUS_BITS_32,1, data);
   dynClear(registersDP);
   dynAppend(registersDP, Tell1Name + ".PP0.CMNMON.ConstantReg"); 
   
   if(fwCcpc_read(registersDP, data, statusG))
     status = 0;
   else
     status = -1;
   
   if(status == 0){
      dummy = fwTell1_BitsFromString(fwCcpc_convertByteToHex(data[1]),7,4);
      //DebugTN("***********dummy:");
      //DebugN(dummy);      
      if(dummy == "1"){
         TypeOfBoard = "VELO";
      }   
      else if(dummy == "2"){
         TypeOfBoard = "ST";
      }
      else if(dummy == "3"){
         TypeOfBoard = "OT";
      }   
      else if(dummy == "4"){
         TypeOfBoard = "EHCAL";
      }
      else if(dummy == "5"){
         TypeOfBoard = "MUON";
      }
      else if(dummy == "6"){
         TypeOfBoard = "L0MUON";
      }
      else if(dummy == "7"){
         TypeOfBoard = "L0DU";
      } 
      else if(dummy == "8"){
         TypeOfBoard = "L0PUS";
      }
      else if(dummy == "9"){
         TypeOfBoard = "RICH";
      }
       else if(dummy == "A"){
         TypeOfBoard = "PSSPD";
         }
       else if(dummy =="B"){
          TypeOfBoard = "L0CAL";          
           }
         else if(dummy ="C"){
         TypeOfBoard = "BCM";             
           }
      else {
         TypeOfBoard = "NOT_DEF";
         status = 1;
      }     
   } else{
        DebugTN("Warning: LBUSRead failed!");
        TypeOfBoard = "ERR";
   }
   //DebugTN(TypeOfBoard);
   return status;
}




/** This function sets bit Nr.1 in the MDIO-Control register (0x683) of the MAC chip to enable/disable the austoscan feature. 
 *  The flag 'true' (=default) enables the autoscan, 'false' will disable the autoscan.
 *  It checks if the WriteRead function terminated successfully and compares the written bit with the value that was read back.
 *
 *  \param [in] Tell1Name The name of the hardware (Tell1 board).
 *  \param [in] pFlag Used to enable or disable autoscan feature of MAC chip.
 *  \return \e int the errorcode of this library: 0 if no error occured
 *                                               -1 register could not be read or written
 *                                                1 if the bit which is set does not match with the bit read back from register
 *
 *  \b Example:
 *  \code
 *  (...)
 *  if(fwTell1_enableMDIOCtrl("myTell1",true) == 0)
 *     DebugN("MDIO Control enabled without error!");
 *  (...)
 *  \endcode
 *
 *  \n
 *  \n
*/

int fwTell1_enableMDIOCtrl(string Tell1Name, bool pFlag = true){
   int status = 0;
   dyn_dyn_char mask;
   dyn_dyn_char dataRead;
   dyn_char maskS;
   dyn_char dataS;
   int dummy;
   int bit; 
   dyn_dyn_char data;  
   dyn_int statusG;
   dyn_string registersDP;
   dynClear(mask);
   dynClear(data);
   dynClear(registersDP);
   dynAppend(registersDP, Tell1Name + ".GBE.MDIOControl"); 
   
   //int address = fwCcpc_convertHexToDec("683");
   if(pFlag){
      dataS = fwCcpc_convertHexToByte("00000002");
      maskS = fwCcpc_convertHexToByte("00000002");
   }
   else{
      dyn_char dataS = fwCcpc_convertHexToByte("00000000");
      dyn_char maskS = fwCcpc_convertHexToByte("00000002");
   }  

   dynAppend(data, dataS);
   dynAppend(mask, maskS);
 
   if(fwCcpc_writeRead(registersDP, data, mask, dataRead, statusG))
     status = 0;
   else{
     DebugN("*** Error: Failed write/read operation! ***");
     return -1;
   }
     
   dummy = fwCcpc_convertByteToDec(dataRead[1]);
   bit = fwTell1_BitsFromInt(dummy,1,1);   
   DebugTN("MDIO Autoscan - Bit value: " + bit);
   //DebugN(bit); 

   if(pFlag){
      if(bit != 1){
         status = 1;
         DebugTN("Error when reading back register - MDIO Autoscan may be not set!");
      }
   }else{ 
      if(bit != 0){
        status = 1;
        DebugTN("Error when reading back register - MDIO Autoscan may be not disabled!");
      }
   }
   return status;
}





/** Checks if Triggers are sent to Tell1 board. Accessing the board e.g. applying recipes should be prohibited then.
 *
 *  \param [in]  Tell1Name The name of the hardware (Tell1 board).
 *  \return \e int Return 1 if triggers are applied and 0 if not. -1 if an error occured.
 *
 *  \b Example:
 *
 *  \code
 *  (...)
 *  if(!fwTell1_chkTriggers("myTell1Board")
 *     fwHw_applyRecipe(DeviceUnitDP_noDist, regs, data);
 *  (...)
 *  \endcode
 *
 *  \n
 *  \n
*/


int fwTell1_chkTriggers(string Tell1Name){
  
   dyn_dyn_char data;  
   dyn_int status;
   dyn_string registers;
   dyn_string registersDP;
 
   int dummy;
   
   //dynClear(registers);
   dynClear(registersDP);   
  // dynAppend(registers, "SL.CMNMON.RealRateReg1.readings");   
   
   //for(int i = 1; i <= dynlen(registers); i++)
   dynAppend(registersDP, Tell1Name + "." + "SL.CMNMON.RealRateReg1");   
   
   if(fwCcpc_read(registersDP, data, status)){  
   
      dummy = fwCcpc_convertByteToDec(data[1]);   
      if (dummy != 0){
        // setValue("attention1","visible",true);
        // setValue("daqStat","visible",true);
        // setValue("apply","enabled",false);
        return 1;
      }
      else{
        //setValue("attention1","visible",false);
        //setValue("daqStat","visible",false);
        //setValue("apply","enabled",true);
        return 0; 
      }
 
   }
   else   
      return -1;
}




  
/** This function disables the GBE ports on the SL (no data sent) and 
 *  sets ECS as trigger source (no TTC triggers accepted). 
 * 
 *  \param [in]  Tell1Name The name of the hardware (Tell1 board).
 *  \return \e int the errorcode of this library: 0 if no error occured
 *                                               -1 ECS error
 *  \b Example:
 *
 *  \code
 *  (...)
 *  if(!fwTell1_disableBoard("myTell1Board")
 *     DebugN("Board is disabled - no TTC triggers accepted and no data can be sent from GBE!"):
 *  (...)
 *  \endcode
 *
 *  \n
 *  \n
*/

 

int fwTell1_disableBoard(string Tell1Name) 
{  
   dyn_dyn_char data, mask;  
   dyn_int statusG;
   dyn_string registers;
   dyn_string registersDP;
   dyn_int dummy;
   dyn_int limit;
   data[1]  = fwCcpc_convertHexToByte("00000000");
   mask[1]= fwCcpc_convertHexToByte("00700000");  
   data[2]  = fwCcpc_convertHexToByte("00000000");
   mask[2] = fwCcpc_convertHexToByte("000000F0");
   
   dynClear(registers);
   dynClear(registersDP);
   dynAppend(registers, "SL.CMNCTRL.SLCtrlReg0");  
   dynAppend(registers, "SL.CMNCTRL.Spi3TxCtrlReg");  

  for(int i = 1; i <= dynlen(registers); i++)
     dynAppend(registersDP, Tell1Name + "." + registers[i]);   
   
  if(!fwCcpc_write(registersDP, data, mask, statusG, true)) 
    return -1;
  return 0;
  
  //address = fwCcpc_convertHexToDec("1000010");     
  //address = fwCcpc_convertHexToDec("1000004");   
} 
 
 
 


/** Resets the sample clock circuit and reinitializes all ADC clock phases. This function has to be called after
 *  configuring the Tell1 board. If not the data from the ARx boards is not sampled correctly. The function is part
 *  of the FSM.
 *
 *  \param [in]  Tell1Name The name of the hardware (Tell1 board).
 *  \return \e int Return 1 if triggers are applied and 0 if not. -1 if an error occured.
 *
 *  \b Example:
 *
 *  \code
 *  (...)
 *  if(!fwTell1_resetVeloSampleClockPhase("myTell1Board")
 *     DebugN("ARx ready for taking data!"):
 *  (...)
 *  \endcode
 *
 *  \n
 *  \n
*/


int fwTell1_resetVeloSampleClockPhase(string Tell1Name){
   dyn_dyn_char data;
   dyn_dyn_char mask;  
   dyn_int statusG;
   dyn_string registersDP;
   dyn_char dataS = fwCcpc_convertHexToByte("00000001");
   dyn_char maskS = fwCcpc_convertHexToByte("00000001");
   int dummy,i;
   
   dynClear(registersDP); 
   dynClear(mask);
   dynClear(data);  
  
   dynAppend(registersDP, Tell1Name + "." + "PP0.CMNCTRL.PPResetReg");  
   dynAppend(registersDP, Tell1Name + "." + "PP1.CMNCTRL.PPResetReg");   
   dynAppend(registersDP, Tell1Name + "." + "PP2.CMNCTRL.PPResetReg");   
   dynAppend(registersDP, Tell1Name + "." + "PP3.CMNCTRL.PPResetReg");    
   
   for(i = 1; i <= 4; i++){
      dynAppend(data, dataS);
      dynAppend(mask, maskS);
   }
   if(fwCcpc_write(registersDP, data, mask, statusG))
     return 0;
   else{
     DebugN("*** Error: Failed write operation! ***");
     return -1;
   }
   
}



/** @} */







// ======================================
//  COMMON CCPC FUNCTIONs
// ======================================

/** @defgroup SectionCCPC CCPC Functions
 *  The input parameter of these functions will always be the CCPC name (name of the server related to the Tell1 board).
 *  These functions will work even when the board's registers are not subsribed yet!
 *  Most of these functions (e.g. the initialisations) are linking directly to the tell1lib provided with the C-code 
 *  command line tools.(hard-coded sequences to avoid timing and sequencing problems)
 *  @{
*/



/** This function loads the FPGA from EPC16 (just for TELL1).
 *
 *  \param [in] pCCPC name of the CCPC.
 *  \return \e int the errorcode of this library: 0 if no error occured
 *                                               -1 ECS error
 *
 *  \b Example:
 *
 *  \code
 *  (...)
 *  if (fwTell1_EPC16ToFPGA("pclbcecs03") == 0)   // load the FPGA from EPC16
 *     (...)
 *  else
 *     DebugTN("Error found!");
 *  (...)
 *  \endcode
 *
 *  \n
 *  \n
*/
int fwTell1_EPC16ToFPGA(string pCCPC)
{

	if (_fwCcpc_commonOperation(pCCPC, "5", true) == "-1")
		return -1;		// ~OK
	else
		return 0;		// OK	
		
}








 
//*******************************************************************
// 
//   functions that are calling hard-coded sequences on the server 
//   this is needed for e.g. configuring the Tell1 board 
 
 
 
 
/** This function initializes the glue-card  
 * 
 *  \param [in] pCCPC name of the CCPC. 
 *  \return \e int the errorcode of this library: 0 if no error occured 
 *                                               -1 ECS error
 * 
 *  \b Example: 
 * 
 *  \code 
 *  (...) 
 *  if (fwTell1_iniGlueCard("pclbcecs03") == 0)  
 *     (...) 
 *  else 
 *     DebugTN("Error found!"); 
 *  (...) 
 *  \endcode 
 * 
 *  \n 
 *  \n 
*/ 
int fwTell1_iniGlueCard(string pCCPC) 
{ 
 
	if (_fwCcpc_commonOperation(pCCPC, "20", true) == "-1") 
		return -1;		// ~OK 
	else 
		return 0;		// OK	 
		 
} 
 
 
 
 
 
 
 
 
 
 
/** This function resets the Tell1 board by pulling some GPIO lines  
 * 
 *  \param [in] pCCPC name of the CCPC. 
 *  \return \e int the errorcode of this library: 0 if no error occured 
 *                                               -1 ECS error
 * 
 *  \b Example: 
 * 
 *  \code 
 *  (...) 
 *  if (fwTell1_resetTell1("pclbcecs03") == 0)  
 *     (...) 
 *  else 
 *     DebugTN("Error found!"); 
 *  (...) 
 *  \endcode 
 * 
 *  \n 
 *  \n 
*/ 
int fwTell1_resetTell1(string pCCPC) 
{ 
 
	if (_fwCcpc_commonOperation(pCCPC, "21", true) == "-1") 
		return -1;		// ~OK 
	else 
		return 0;		// OK	 
		 
} 
 
 
 
 
 
 
 
/** This function sets a bit in SL_RESET_REG to reset the TELL1 board  
 * 
 *  \param [in] pCCPC name of the CCPC. 
 *  \return \e int the errorcode of this library: 0 if no error occured 
 *                                               -1 ECS error
 * 
 *  \b Example: 
 * 
 *  \code 
 *  (...) 
 *  if (fwTell1_resetTell1Soft("pclbcecs03") == 0)  
 *     (...) 
 *  else 
 *     DebugTN("Error found!"); 
 *  (...) 
 *  \endcode 
 * 
 *  \n 
 *  \n 
*/ 
int fwTell1_resetTell1Soft(string pCCPC) 
{ 
 
	if (_fwCcpc_commonOperation(pCCPC, "22", true) == "-1") 
		return -1;		// ~OK 
	else 
		return 0;		// OK	 
		 
} 
 
 
 
 
 
 


 
 
 
/** This function initializes the Gigabit Ethernet card 
 * 
 *  \param [in] pCCPC name of the CCPC. 
 *  \return \e int the errorcode of this library: 0 if no error occured 
 *                                               -1 ECS error
 * 
 *  \b Example: 
 * 
 *  \code 
 *  (...) 
 *  if (fwTell1_iniGBE("pclbcecs03") == 0)  
 *     (...) 
 *  else 
 *     DebugTN("Error found!"); 
 *  (...) 
 *  \endcode 
 * 
 *  \n 
 *  \n 
*/ 
int fwTell1_iniGBE(string pCCPC) 
{ 
 
	if (_fwCcpc_commonOperation(pCCPC, "23", true) == "-1") 
		return -1;		// ~OK 
	else 
		return 0;		// OK	 
		 
} 
 
 
 
 
 
 
 
/** This function clears the MEP buffer  
 * 
 *  \param [in] pCCPC name of the CCPC. 
 *  \return \e int the errorcode of this library: 0 if no error occured  
 *                                               -1 ECS error
 * 
 *  \b Example: 
 * 
 *  \code 
 *  (...) 
 *  if (fwTell1_clearMEP("pclbcecs03") == 0)  
 *     (...) 
 *  else 
 *     DebugTN("Error found!"); 
 *  (...) 
 *  \endcode 
 * 
 *  \n 
 *  \n 
*/ 
int fwTell1_clearMEP(string pCCPC) 
{ 
 
	if (_fwCcpc_commonOperation(pCCPC, "24", true) == "-1") 
		return -1;		// ~OK 
	else 
		return 0;		// OK	 
		 
} 
 
 
 
 
 
 
 
 
/** This function initializes the complete Tell1 board. It calls all the necessary hardcoded sequences 
 *  directly on the server. For propper configuration it is necessary to call this function before the 
 *  recipes are applied.  
 * 
 *  \param [in] pCCPC name of the CCPC. 
 *  \return \e int the errorcode of this library: 0 if no error occured 
 *                                               -1 ECS error
 * 
 *  \b Example: 
 * 
 *  \code 
 *  (...) 
 *  if (fwTell1_iniTell1("pclbcecs03") == 0)  
 *     (...) 
 *  else 
 *     DebugTN("Error found!"); 
 *  (...) 
 *  \endcode 
 * 
 *  \n 
 *  \n 
*/ 
int fwTell1_iniTell1(string pCCPC) 
{ 
 
	if (_fwCcpc_commonOperation(pCCPC, "25", true) == "-1") 
		return -1;		// ~OK 
	else 
		return 0;		// OK	 
		 
} 
 
 





/** @} */





// ======================================
//  FUNCTIONALITY CHECK FUNCTIONs
// ======================================

/** @defgroup SectionCheck Functionality-Check Functions
 *  List of functions to evaluate the status of the Tell1 boards. 
 *  Used for testing, whether they are operational and functional. 
 *  The input parameter of these functions will always be the hardware name (datapoint name).
 *  The functions will just work, if the board is subscribed!
 *  These functions are used for the board evaluation on the device unit panels.
 *  @{
*/


/** Checks if the CCPC is running by checking the invalid bit of the 'common.operationStatus' service.
 *
 *  \param [in] string Tell1Name the name of the hardware (Tell1 board).
 *  \param [out] string CCPC name of the CCPC found for Tell1 board.
 *  \param [out] dyn_string Version the version of the server running - first element: version, second element: release, third element: package.
 *  \return \e int the errorcode of this library: 0 no error occured and server is running 
 *                                               -1 error related to ECS interaction
 *                                                1 server is down
 *
 *  \b Example:
 *  
 *  \code
 *  (...)
 *  string CCPC;
 *  dyn_int Version;
 *     if(fwTell1_chkCCPC("myTell1", CCPC, Version) == 0)
 *        DebugN("Server found!");
 *        DebugN("CCPC: " + CCPC);
 *        DebugN("Version: " + Version);
 *  (...)
 *  \endcode
 *
 *  \n
 *  \n
 *  
*/

int fwTell1_chkCCPC(string Tell1Name, string &CCPC, dyn_int &Version){
 
   //dyn_string tmp0;
   //dyn_string tmp1;
   //dynClear(tmp0);
   //dynClear(tmp1);
   //int i;	
   //CcpcOn = 0
   int error = 0;
   bool invalidBit;
   string dummy, dist, pCCPC;
   int a, b, c; 
   dyn_string splitDP;
   //string DeviceUnitDP;
//a temporary hack to set the invalid bit - later service is published immediately on start up
   
   //get the ccpc of the Tell1
   dummy = Tell1Name + ".settings";
   if(dpGet(dummy,CCPC)){
     DebugN("Failed doing dpGet!");
     return -1;
   }
   
   splitDP = strsplit(Tell1Name, ":");
		//DebugTN(splitCCPC);
		if(dynlen(splitDP) == 2){
			pCCPC = splitDP[1] + ":" + strtoupper(FWCCPC_PREFIX) + CCPC;
                        CCPC = splitDP[1] + ":" + CCPC;
			DebugN(pCCPC);
                        DebugN(CCPC);
		}
		else if(dynlen(splitDP) == 1){
			pCCPC = strtoupper(FWCCPC_PREFIX) + CCPC;
			DebugN(pCCPC);
		}
		else
			DebugN("****WARNING:  Do not use semicolons in your name except for separation of system name and ccpc name!!!! ****");
			
  
   

   DebugTN("Check if server is connected!");
   if(dpGet(pCCPC+".common.operationStatus:_original.._invalid",invalidBit)){
     DebugN("Failed getting invalidBit!");
     return -1;
   }
   
   
   //DebugTN(invalidBit);
   if(invalidBit == true){
      DebugTN("Server "+CCPC+" is down!");
      return 1;
   } else{
      DebugTN("Server "+CCPC+" found!");
      //return 0;
   }

   
      
    if(!fwCcpc_getServerVersion(CCPC, a, b, c)){
     DebugN("Failed getting server version!");
     return -1; 
   }
  
   //DebugN(a, b, c);
   dynClear(Version);
   dynAppend(Version,a);
   dynAppend(Version,b);
   dynAppend(Version,c);
   //DebugN(Version);
    DebugN("Server Version: " + Version);
   if(a==0 || b==0 || c==0){
     DebugN("Server version not correctly identified!");
     return -1;
   }

   return 0;
}
   


/*
   DebugTN("chkCCPC - looking for command:"); 
   if (isFunctionDefined("fwDim_createConfig")){
         fwDim_findCommands(FWCCPC_CONFIG_NAME, FWCCPC_PREFIX  + CCPC + "/CmndCommonOperation", tmp0, tmp1);	
                  DebugN(FWCCPC_CONFIG_NAME, FWCCPC_PREFIX + CCPC + "/CmndCommonOperation");

      if(dynlen(tmp0) == 1){
         DebugTN("ccpc found!");
         return 0;}
      else{
	 DebugTN("ccpc not found!");
         return 1;}	
    }
*/






/** Checks if the .cfg file was written propperly by evaluating a part of a certain register.
 *  The function is not supported for distributed systems.
 *
 *  \param [in] CCPC name of the credit card PC.
 *  \return \e int the errorcode of this library: 0 if no error occured
 *                                               -1 LBUS access failed
 *                                                positive value (5-bit number): one bit per failed FPGA
 *
 *  \b Example:
 *
 *  \code
 *  (...)
 *  string ccpc = "pctell07";
 *
 *  if(fwTell1_chkCFG(ccpc))
 *     DebugTN("FPGAs are not configured propperly!");
 *  (...)
 *  \endcode
 *
 *  \n
 *  \n
*/

int fwTell1_chkCFG(string CCPC){
  //int fwTell1_chkCFG(string Tell1Name){
   dyn_int status;
   int error = 0;
   dyn_dyn_char data;
   int address;
   int myhex;
   //dyn_string registersDP;
   
   //dynClear(registersDP);   
   //dynAppend(registersDP, Tell1Name + "." + "SL.SLCtrlReg0");   
   
   //if(!fwCcpc_read(registersDP, data, status)){  
   address = fwCcpc_convertHexToDec("1000004");   
   if(fwCcpc_LBUSRead(CCPC, address, FWCCPC_LBUS_BITS_32, 1, data)){
      DebugTN("Warning: LBUSRead failed!");
      return -1;
   }      

   myhex = fwTell1_BitsFromInt(fwCcpc_convertByteToDec(data[1]), 15, 11);
   DebugTN("Check CFG: Returned 5-bit number (hex):");
   DebugN(fwCcpc_convertDecToHex(myhex));
 
   if( myhex == 0x1F){
      //FPGA init OK
      return 0;
   }else{
      //FPGA init NOT OK
      return myhex;
   }
}







/** Checks if GBE is operational - it checks if ports are enabled and not plugged, which would lead to a data loss.
 *  And it checks if all ports are disabled.
 *
 *  \param [in] Tell1Name The name of the hardware (Tell1 board).
 *  \return \e int the errorcode of this library: 0 if no error occured
 *                                               -1 failed reading the registers
 *                                                1 Port0 not plugged but enabled (data loss!)
 *                                                2 Port1 not plugged but enabled (data loss!)
 *                                                4 Port2 not plugged but enabled (data loss!)
 *                                                8 Port3 not plugged but enabled (data loss!)
 *                                                up to 15 any combination of the error code below
 *                                                e.g. 3 --> Port0 & Port1 are not plugged but enabled
 *                                                16 all ports are disabled on SL (no dataflow!)
 *
 *  \b Example:
 *
 *  \code
 *  (...)
 *  if(fwTell1_chkGBE("myTell1") == 16)
 *     DebugTN("WARNING: All ports are disabled on SyncLink!");
 *  (...)
 *  \endcode
 *
 *  \n
 *  \n
*/


int fwTell1_chkGBE(string Tell1Name){
   int status = 0;
   int status2 = 0;
   
   dyn_dyn_char data;  
   dyn_int statusG;
   dyn_string registers;
   dyn_string registersDP;
  
   fwTell1_enableMDIOCtrl(Tell1Name); 
      
   dynClear(registers);
   dynClear(registersDP); 
   dynAppend(registers, "SL.CMNCTRL.Spi3TxCtrlReg");  
   dynAppend(registers, "GBE.PORT0.PHYStatus");  
   dynAppend(registers, "GBE.PORT1.PHYStatus");  
   dynAppend(registers, "GBE.PORT2.PHYStatus");  
   dynAppend(registers, "GBE.PORT3.PHYStatus"); 
   
   for(int i = 1; i <= dynlen(registers); i++){
      dynAppend(registersDP, Tell1Name + "." + registers[i]); 
   }
   
   if(fwCcpc_read(registersDP, data, statusG)){
      dyn_int dummy;
      for(int i = 1; i<= dynlen(data); i++)
         dummy[i] = fwCcpc_convertByteToDec(data[i]);
  
      for(int j = 1; j<=4;j++)  
        if((fwTell1_BitsFromInt(dummy[j+1],2,2) != 1)  && (fwTell1_BitsFromInt(dummy[1],3+j,3+j) == 1)){
        if(j == 1)  status += 1;
        if(j == 2)  status += 2;
        if(j == 3)  status += 4;
        if(j == 4)  status += 8; 
        
        DebugTN(j , status);    
          
        }
      //DATA LOSS, the port is not plugged, but the link between sl and gbe is enabled

      for(int j = 1; j<=4;j++)  {
        if((fwTell1_BitsFromInt(dummy[1],3+j,3+j) == 0)){
           status2 += 1;
      }
    }
      if(status2 == 4){
       status = 16;
     }
     return status;
 }else{
       DebugN("Warning: Could not read registers!");
       return -1; 
   }
}




/** Checks if TTC is operational (fibre is plugged). If fibre is unplugged the I2C interface is pertubated. 
 *  This function is reading 3-times from the status register. 
 *  In addition the last two bits from the status register are checked.
 *
 *  \param [in] Tell1Name The name of the hardware (Tell1 board).
 *  \return \e int the errorcode of this library: 0 if no error occured
 *                                                1  clock or data recovery circuit not locked on incoming data stream
 *                                                2  high-resolution phase shifters not working propperly!
 *                                                3  both bits are set
 *                                               -1  I2C-read of status register failed (minimum once out of three trials)
 *
 *  \b Example:
 *
 *  \code
 *  (...)
 *  if(fwTell1_chkTTC("myTell1"))
 *     DebugTN("WARNING: Fibre might not be plugged on TTC!"
 *  (...)
 *  \endcode
 *
 *  \n
 *  \n
*/

int fwTell1_chkTTC(string Tell1Name){
   int i;
   int error = 0;
   dyn_dyn_char data;
   int dllReady;
   //indicates that the high-resolution phase shifters are working propperly
   int pllReady;
   //indicates that the clock and data recovery circuit is locked on the incoming data stream
   dyn_string registersDP;
   dyn_int statusG;
   dynClear(registersDP);
   dynAppend(registersDP, Tell1Name + ".TTCRX.Status"); 
    
   //int address = fwCcpc_convertHexToDec("58");
   //int address = 0x58;
   for(i=1; i <= 3; i++){
     // if(fwCcpc_I2CReadSub(CCPC, 0, address, 22, 1, -1, FWCCPC_I2C_SEPARATED, 0, data))
      if(!fwCcpc_read(registersDP, data, statusG)){  
         DebugN("Warning: TTCRead failed!");
         error += 1;
      }     
   }
   if(error){
      DebugTN("Could not read I2C interface - check clock and fibre of TTCRx!");
      return -1;
   }
      
   error = 0;
   pllReady = fwTell1_BitsFromInt(fwCcpc_convertByteToDec(data[1]), 7, 7);
   dllReady = fwTell1_BitsFromInt(fwCcpc_convertByteToDec(data[1]), 6, 6);
   if((pllReady == 0)){
      DebugTN("TTC: Clock or data recovery circuit not locked on incoming data stream!");
      error += 1;
   }
   if(dllReady==0){
      DebugTN("TTC: High-resolution phase shifters not working propperly!");
      error += 2;
   }
   return error;
}





/** Checks if ORx is operational - e.g. link is enabled and not in state error.
 *
 *  \param [in] Tell1Name The name of the hardware (Tell1 board).
 *  \return \e int the errorcode of this library: 0 if no error occured
 *                                               -1 registers could not be read
 *                                                1 at least one link is enabled and in state error
 *                                                2 all links are disabled
 *
 *  \b Example:
 *  \code
 *  (...)
 *  if(fwTell1_chkMemory("myTell1") > 0)
 *     DebugTN("WARNING: Memory Overflow!");
 *  (...)
 *  \endcode
 *
 *  \n
 *  \n
*/


int fwTell1_chkORx(string Tell1Name){
   int status = 0;
   int disableLink = 0; 

   dyn_dyn_char data;  
   dyn_int statusG;
   dyn_string registers;
   dyn_string registersDP;
  
   string dummy1;
   string dummy2;
   string dummy3;
   
   dynClear(registers);
   dynClear(registersDP);
   
   dynAppend(registers, "PP0.ORX.link0");  
   dynAppend(registers, "PP0.ORX.link1");  
   dynAppend(registers, "PP0.ORX.link2");  
   dynAppend(registers, "PP0.ORX.link3");  
   dynAppend(registers, "PP0.ORX.link4");  
   dynAppend(registers, "PP0.ORX.link5");  
   dynAppend(registers, "PP1.ORX.link0");  
   dynAppend(registers, "PP1.ORX.link1");  
   dynAppend(registers, "PP1.ORX.link2");  
   dynAppend(registers, "PP1.ORX.link3");  
   dynAppend(registers, "PP1.ORX.link4");  
   dynAppend(registers, "PP1.ORX.link5");  
   dynAppend(registers, "PP2.ORX.link0");  
   dynAppend(registers, "PP2.ORX.link1");  
   dynAppend(registers, "PP2.ORX.link2");  
   dynAppend(registers, "PP2.ORX.link3");  
   dynAppend(registers, "PP2.ORX.link4");  
   dynAppend(registers, "PP2.ORX.link5");  
   dynAppend(registers, "PP3.ORX.link0");  
   dynAppend(registers, "PP3.ORX.link1");  
   dynAppend(registers, "PP3.ORX.link2");  
   dynAppend(registers, "PP3.ORX.link3");  
   dynAppend(registers, "PP3.ORX.link4");  
   dynAppend(registers, "PP3.ORX.link5"); 
   
    
   for(int i = 1; i <= dynlen(registers); i++){
      dynAppend(registersDP, Tell1Name + "." + registers[i]); 
   }
   
   if(fwCcpc_read(registersDP, data, statusG)){

   for(int i = 1; i <=24; i++){
   //fwCcpc_read ....
     
    
    if (fwTell1_BitsFromInt(fwCcpc_convertByteToDec(data[i]), 23, 23) != "0"){  
           disableLink++;
	  }else if(fwTell1_BitsFromInt(fwCcpc_convertByteToDec(data[i]), 14, 14) == "0" && fwTell1_BitsFromInt(fwCcpc_convertByteToDec(data[i]), 12, 12) == "0"){  
           status++;
    }
 
   }
   
   DebugTN("   "+disableLink +" ORx link disabled");
   int en = 24 - disableLink;
   DebugTN("   "+en +" ORx link enabled and "+status+" are(is) in error state");
   if(disableLink ==  24)     
      return 2;
   if(status !=  0)     
      return 1;
   
   
   return 0; 
   } else{
      
       DebugN("Warning: ORxRead failed!"); 
   }
   return -1;
}


/** Checks if ARx is operational - e.g. the DACs are connected and are accessible via ECS.
 *
 *  \param [in] Tell1Name The name of the hardware (Tell1 board).
 *  \return \e int the errorcode of this library: 0 if no error occured
 *                                               -1 read access of registers failed
 *                                                1 DAC0 not plugged
 *                                                2 DAC1 not plugged
 *                                                4 DAC2 not plugged
 *                                                8 DAC3 not plugged
 *                                                up to 15 any combination of the error code below
 *                                                e.g. 3 DAC0 & DAC1 not plugged
 *
 *  \b Example:
 *
 *  \code
 *  (...)
 *  if(fwTell1_chkARx("myTell1") > 0)
 *     DebugTN("WARNING: Some DACs may not be plugged!");
 *  (...)
 *  \endcode
 *
 *  \n
 *  \n
*/

int fwTell1_chkARx(string Tell1Name){
   int status = 0; 
   
   dyn_dyn_char data;  
   dyn_int statusG;
   dyn_string registers;
   dyn_string registersDP;
   dyn_int dummy;
   dyn_int limit;

   dynClear(registers);
   dynClear(registersDP);
   
   dynAppend(registers, "ARX.DACPP0.BurstWrite_Data");  //freeze the cnts
   dynAppend(registers, "ARX.DACPP1.BurstWrite_Data");      
   dynAppend(registers, "ARX.DACPP2.BurstWrite_Data");      
   dynAppend(registers, "ARX.DACPP3.BurstWrite_Data");      
       
   for(int i = 1; i <= dynlen(registers); i++){
      dynAppend(registersDP, Tell1Name + "." + registers[i]); 
   }
   
   if(!fwCcpc_read(registersDP, data, statusG))
     status = -1;
   
   if(statusG[1] == 4){
      status += 1; 
      DebugTN("DAC 0 not plugged");
   }
   if(statusG[2] == 4){
      status += 2;
      DebugTN("DAC 1 not plugged");
   }
   if(statusG[3] == 4){
      status += 4;
      DebugTN("DAC 2 not plugged");
   }
   if(statusG[4] == 4) {
      status += 8; 
      DebugTN("DAC 3 not plugged");
   } 
   return status;
}




/** Checks if processing on Tell1 board is ok  - the counter registers are frozen and compared.
 *
 *  \param [in] Tell1Name The name of the hardware (Tell1 board).
 *  \return \e int the errorcode of this library: 0 if no error occured (all 4 counters equal)
 *                                               -1 registers could not be read
 *                                                1 Counters are not equal
 *
 *  \b Example:
 *  \code
 *  (...)
 *  if(fwTell1_chkProcessing("myTell1") == 1)
 *     DebugTN("WARNING: Counter mismatch on FPGAs - processing might be corrupted! Please check Proc. Mon. panel.");
 *  (...)
 *  \endcode
 *
 *  \n
 *  \n
*/


int fwTell1_chkProcessing(string Tell1Name){
   int status = 0; 
   
   dyn_dyn_char data;  
   dyn_int statusG;
   dyn_string registers;
   dyn_string registersDP;
   dyn_int dummy;
   dyn_int limit;

   dynClear(registers);
   dynClear(registersDP);
   
  dynAppend(registers, "SL.REALTIME.SLTrigCntReg");  //freeze the cnts
  dynAppend(registers, "SL.CMNMON.SLEvtInCntReg0");      
  dynAppend(registers, "SL.CMNMON.SLEvtInCntReg1");      
  dynAppend(registers, "SL.CMNMON.SLEvtInCntReg2");      
  dynAppend(registers, "SL.CMNMON.SLEvtInCntReg3");      
     
  for(int i = 1; i <= dynlen(registers); i++)
     dynAppend(registersDP, Tell1Name + "." + registers[i]);   
   
  if(fwCcpc_read(registersDP, data, statusG)){   
   if(fwCcpc_convertByteToHex(data[2]) == fwCcpc_convertByteToHex(data[3]) && fwCcpc_convertByteToHex(data[4]) == fwCcpc_convertByteToHex(data[5]) && fwCcpc_convertByteToHex(data[2]) == fwCcpc_convertByteToHex(data[4]))
         status = 0;
     else
        status =1; 
  }else{
     DebugN("Warning: Reading registers for chkProcessing failed!");
     status = -1; 
  }
  return status;
}




/** Checks if data flow on Tell1 board is ok  - e.g. whether throttles are arriving or not.
 *  The throttle counter is read twice with a pause of 100ms in between. The two values are checked whether
 *  they are equal or not.
 *
 *  \param [in] Tell1Name The name of the hardware (Tell1 board).
 *  \return \e int the errorcode of this library: 0 if no error occured
 *                                               -1 the counter register could not be read
 *                                                1 throttles arrived (counter incremented within 100ms)
 *
 *  \b Example:
 *
 *  \code
 *  (...)
 *  if(fwTell1_chkFlow("myTell1") > 0)
 *     DebugTN("WARNING: Throttles arriving!");
 *  (...)
 *  \endcode
 *
 *  \n
 *  \n
*/



int fwTell1_chkFlow(string Tell1Name){
   int status = 0;
   //chk if throttle ae coming
   dyn_dyn_char data1, data2;  
   dyn_int statusG;
   dyn_string registers;
   dyn_string registersDP;
   dynClear(registers);
   dynClear(registersDP);
   dynAppend(registers, "SL.CMNMON.ThroCntReg0");  
   
   for(int i = 1; i <= dynlen(registers); i++){
      dynAppend(registersDP, Tell1Name + "." + registers[i]); 
   }
   
  if(fwCcpc_read(registersDP, data1, statusG)){  
     delay(0,100);
     if(fwCcpc_read(registersDP, data2, statusG)){   
        if(fwCcpc_convertByteToHex(data1[1]) != fwCcpc_convertByteToHex(data2[1]))
           status = 1;
        return status;
     }else{
        DebugN("Warning: failed reading throttle counter register! (SL.CMNMON.ThroCntReg0)");
        return -1; 
     }
  }else{
    DebugN("Warning: failed reading throttle counter register! (SL.CMNMON.ThroCntReg0)");
    return -1; 
  }
}



/** Checks if memory on Tell1 board is ok  - memory overflow occurs if the number of used bits exceeds a certain limit
 *
 *  \param [in] string Tell1Name The name of the hardware (Tell1 board).
 *  \return \e int the errorcode of this library: 0 if no error occured
 *                                               -1 registers could not be read
 *                                               >0 number of registers where memory overflow occured
 *  \b Example:
 *  \code
 *  (...)
 *  if(fwTell1_chkMemory("myTell1") > 0)
 *     DebugTN("WARNING: Memory Overflow!");
 *  (...)
 *  \endcode
 *
 *  \n
 *  \n
*/



int fwTell1_chkMemory(string Tell1Name){
   int status = 0;
  
   dyn_dyn_char data;  
   dyn_int statusG;
   dyn_string registers;
   dyn_string registersDP;
   dyn_int dummy;
   dyn_int limit;
 
   dynClear(registers);
   dynClear(registersDP);   
  dynAppend(registers, "SL.CMNMON.SLTriggerFifoUsedReg0");  
  dynAppend(registers, "SL.CMNMON.SLTriggerFifoUsedReg1");      
  dynAppend(registers, "SL.REALTIME.TrigInfoFifoMonReg0");  
  dynAppend(registers, "SL.REALTIME.TrigInfoFifoMonReg1");   
  dynAppend(registers, "SL.CMNMON.MepMaxUsageReg");  
  dynAppend(registers, "SL.CMNMON.SLMaxUseReg");  
  dynAppend(registers, "SL.CMNMON.FramerMaxUseReg");  
  dynAppend(registers, "SL.CMNMON.MepGt16KCntReg");     
  dynAppend(registers, "PP0.CMNMON.MemMaxUsageReg0");  
  dynAppend(registers, "PP1.CMNMON.MemMaxUsageReg0");  
  dynAppend(registers, "PP2.CMNMON.MemMaxUsageReg0");
  dynAppend(registers, "PP3.CMNMON.MemMaxUsageReg0");  
  dynAppend(registers, "PP0.CMNMON.MemMaxUsageReg1");
  dynAppend(registers, "PP1.CMNMON.MemMaxUsageReg1");  
  dynAppend(registers, "PP2.CMNMON.MemMaxUsageReg1");  
  dynAppend(registers, "PP3.CMNMON.MemMaxUsageReg1");  

   for(int i = 1; i <= dynlen(registers); i++){
      dynAppend(registersDP, Tell1Name + "." + registers[i]); 
   }
   
   if(fwCcpc_read(registersDP, data, statusG)){  
      
  for(int i = 1; i <= dynlen(data); i++)
    dummy[i] = fwCcpc_convertByteToDec(data[i]);
   
  limit[1] = fwCcpc_convertHexToDec("1F");   //5bit
  limit[2] = fwCcpc_convertHexToDec("3F");    //7bit 
  limit[3] = fwCcpc_convertHexToDec("FF");    //8bit  
  limit[4] = fwCcpc_convertHexToDec("3FF");    //10bit
  limit[5] = fwCcpc_convertHexToDec("7FF");     //11bit
  limit[6] = fwCcpc_convertHexToDec("FFFF");    //16bit  
  limit[7] = fwCcpc_convertHexToDec("7FFFF");   //19bit   
  limit[8] = fwCcpc_convertHexToDec("FFFFFFFF");    //32bit   
   
  if(fwTell1_BitsFromInt(dummy[3],31,24) == limit[3])
    status ++;
  if(fwTell1_BitsFromInt(dummy[3],23,16) == limit[3])
    status ++;
  if(fwTell1_BitsFromInt(dummy[3],15,8) == limit[3])
    status ++;
  if(fwTell1_BitsFromInt(dummy[4],31,24) == limit[3])
    status ++;
  if(fwTell1_BitsFromInt(dummy[4],23,16) == limit[3])
    status ++;
  
  for(int j = 1; j<=4; j++){
  if(fwTell1_BitsFromInt(dummy[8+j],31,16) == limit[6])
    status ++;
  if(fwTell1_BitsFromInt(dummy[8+j],15,0) == limit[6])
    status ++;
  if(fwTell1_BitsFromInt(dummy[12+j],29,20) == limit[4])
    status ++;
  if(fwTell1_BitsFromInt(dummy[12+j],19,10) == limit[4])
    status ++;
  if(fwTell1_BitsFromInt(dummy[12+j],9,0) == limit[4])
   status ++;

 }
  
  
  if(fwTell1_BitsFromInt(dummy[5],18,0) == limit[7])
    status ++;
  if(fwTell1_BitsFromInt(dummy[6],9,0) == limit[4])
    status ++;
  if(fwTell1_BitsFromInt(dummy[6],20,10) == limit[5])
    status ++;
  if(fwTell1_BitsFromInt(dummy[6],27,21) == limit[2])
    status ++;
  if(fwTell1_BitsFromInt(dummy[7],4,0) == limit[1])
    status ++;
  if(fwTell1_BitsFromInt(dummy[7],11,5) == limit[2])
    status ++;
  if(fwTell1_BitsFromInt(dummy[7],19,12) == limit[3])
    status ++;
  if(dummy[8]!=0)
    status ++;

  
   return status;
   
   
   }else{  
       DebugN("Warning: reading registers failed!");
       return -1; 
   }
}



/** Checks if the .cfg file was written propperly by evaluating a part of a certain register.
 *  This function is supported for distributed systems.
 *
 *  \param [in] Tell1Name name of the hardware (datapoint of Tell1 board).
 *  \return \e int the errorcode of this library: 0 if no error occured
 *                                               -1 LBUS access failed
 *                                                positive value (5-bit number): one bit per failed FPGA
 *
 *  \b Example:
 *
 *  \code
 *  (...)
 *  string ccpc = "myTell1";
 *
 *  if(fwTell1_chkCFG_dist(Tell1Name))
 *     DebugTN("FPGAs are not configured propperly!");
 *  (...)
 *  \endcode
 *
 *  \n
 *  \n
*/

int fwTell1_chkCFG_dist(string Tell1Name){
  //int fwTell1_chkCFG(string Tell1Name){
   dyn_int status;
   int error = 0;
   dyn_dyn_char data;
   int address;
   int myhex;
   dyn_string registersDP;
   
   dynClear(registersDP);   
   dynAppend(registersDP, Tell1Name + "." + "SL.CMNCTRL.SLCtrlReg0");   
   
   if(!fwCcpc_read(registersDP, data, status)){  
   //address = fwCcpc_convertHexToDec("1000004");   
   //if(fwCcpc_LBUSRead(CCPC, address, FWCCPC_LBUS_BITS_32, 1, data)){
      DebugTN("Warning: LBUSRead failed!");
      return -1;
   }      

   myhex = fwTell1_BitsFromInt(fwCcpc_convertByteToDec(data[1]), 15, 11);
   DebugTN("Check CFG: Returned 5-bit number (hex):");
   DebugN(fwCcpc_convertDecToHex(myhex));
 
   if( myhex == 0x1F){
      //FPGA init OK
      return 0;
   }else{
      //FPGA init NOT OK
      return myhex;
   }
}



/** @} */






// ======================================
//  INTERNAL FUNCTIONs (NOT TO BE USED)
// ======================================



// This function is used by Cedric.

string complStr(string imput, int total)
{	
	 string zeros ;
		if(strlen(imput) != total){
	    for(int h = 1 ; h <= total - strlen(imput); h++){
		       
		    zeros = zeros + "0";
		 }
    }
		
	  string output = zeros + imput;
		
		return output;
}



/*
string _fwCcpc_getConfigName()
{

	return FWCCPC_CONFIG_NAME;
	
}




string _fwCcpc_getPrefix()
{

	return FWCCPC_PREFIX;
	
}
*/



//various attempts to improve the assemble function

/*

fwTell1_AssembleAndCreateRecipes(string hardware, dyn_string recipes, string RecipeName, string dpType1, string RecipeType){ 
  //dyn_dyn_string Registers; 
  //dyn_dyn_int Registers_Size; 
  //dyn_dyn_int Registers_Type; 
  //dyn_string  HardwRegisters_ALL, regs;  
  //dyn_int   Types_ALL;
  //dyn_dyn_anytype RecipeData1, RecipeData2,RecipeDataALL; 
  //dyn_dyn_anytype data1, mask1; 
  //dyn_dyn_anytype data2, mask2; 
  //dyn_dyn_anytype dummy; 
  int j,i; 
  int num; 
  bool SameRegistersFlag; 
   
  dyn_string kRegs, kRegsAll, regs;
  dyn_dyn_anytype kData, kDataAll;
  dyn_int kRegSize, kRegType, kRegSizeAll, kRegTypeAll;
  dyn_anytype data1, mask1, dummy; 
  dyn_anytype data2, mask2; 
  
  DebugTN("=========================================================================="); 
  DebugTN("====================ASSEMBLE=THE=DIFFERENT=RECIPES========================");  
  DebugN(recipes);
  if(dynlen(recipes) > 1){     
	fwHw_getRecipe(hardware, recipes[1], kRegsAll, kDataAll);
        //fwTell1_getRegInfo  givs the size and the type of each registers 
        fwTell1_getRegInfo(hardware, kRegsAll, kRegSizeAll, kRegTypeAll); 
        DebugN(kRegsAll); 
 	for(num = 2; num <= dynlen(recipes); num++){ 
	                fwHw_getRecipe(hardware, recipes[num], kRegs, kData); 
                        fwTell1_getRegInfo(hardware, kRegs, kRegSize, kRegType);
                        DebugN(kRegs);
                        DebugN(kRegSize);
                        DebugN(kRegType);
	                if(dynlen(kRegs)>0){ 
	   			for(i = 1; i <= dynlen(kRegs); i++){ 
	   				SameRegistersFlag = false; 
	     				for(j = 1; j<=  dynlen(kRegsAll); j++){	      					  
	                                        if(kRegsAll[j] == kRegs[i]){ 							 
							SameRegistersFlag = true; 
							break ; 
						}               
					} 
					if(SameRegistersFlag){ 
						//fwTell1_updateVar gives back data and mask 
	    					fwTell1_updateVar(kDataAll[j], kRegSizeAll[j], data1, mask1); 
					        fwTell1_updateVar(kData[i], kRegSize[i], data2, mask2);    
						DebugTN("re-create data for register: "+kData[i]); 
						//update of the data and the mask (according to the two old ones) 
                                                dynClear(dummy);
					        fwTell1_writeToVar(data1,mask1,data2,mask2) ; 
						dynAppend(dummy, data1); 
						dynAppend(dummy, mask1);    
                                                kDataAll[i] = dummy;
					}else{ 
                                                dynAppend(kDataAll, kData[i]);
                                                dynAppend(kRegSizeAll, kRegSize[i]);  
                                                dynAppend(kRegTypeAll, kRegType[i]); 
                                                //for the Recipes Type ".SL.blabla" 
					        //dynAppend(kRegAll, substr(kRegs[i],strlen(hardware))); 
                                                dynAppend(kRegsAll, kRegs);
					} 
				}
		}else{ 
 		DebugTN("No registers found in : "+recipes[num]+"!"); 
		} 
	} 
   }else{ 
   DebugTN("Less than two recipes passed as argument!"); 
   } 
   
   
     for(i=1; i<=dynlen(kRegsAll); i++){
    dum = strsplit(kRegsAll[i],":");
            if(dynlen(dum) >= 2)
              kRegsAll[i] = dum[2];
      kRegsAll[i] = substr(kRegsAll[i],strlen(hardware));
          }
   
   
   
   //DebugTN(HardwRegisters_ALL);
   fwHw_createRecipeType(RecipeType,dpType1,kRegsAll);
   DebugN("Registers belonging to recipe type: " + RecipeType);  
   //fwHw_getRecipeTypeInfo(RecipeType, dpType1, regs); 
   DebugN(kRegsAll);    
   fwHw_saveRecipe(hardware, RecipeType, RecipeName, kRegsAll, kRegTypeAll, kDataAll);
   // DebugN(hardware, RecipeType, RecipeName); 
   // DebugN(hardware, RecipeType, RecipeName); 
   DebugTN("=========================END=OF=THE=ASSEMBLY=============================="); 
   DebugTN("=========================================================================="); 
 }

*/



 


   

/*
 
fwTell1_AssembleAndCreateRecipes(string hardware, dyn_string recipes, string RecipeName, string dpType1, string RecipeType){ 

dyn_string AllRegisters;
dyn_int AllRegistersSize;
dyn_string AllRegistersType;   
dyn_dyn_anytype AllRegistersDataCombined;
dyn_dyn_anytype AllRegistersData;
dyn_dyn_anytype AllRegistersMask;

dyn_string registers;
dyn_dyn_anytype data;
dyn_int size;
dyn_string type;

dyn_dyn_anytype  splitdata;
dyn_dyn_anytype  splitmask;
int existing = 0;
int i, j, num;

  DebugTN("=========================================================================="); 
  DebugTN("====================ASSEMBLE=THE=DIFFERENT=RECIPES========================");  
 
   
  if(dynlen(recipes) > 1){                                              //loop over recipes 
 	for(num = 1; num <= dynlen(recipes); num++){ 
            dynClear(registers);
            dynClear(data);
            dynClear(type);
            dynClear(size);
            fwHw_getRecipe(hardware, recipes[num], registers, data);          //load info of recipe's regs   
	   	if(dynlen(registers)>0){             
	     	   fwTell1_getRegInfo(hardware, registers, size, type);     //load info of recipe's regs
	          
	         for(i = 1; i <= dynlen(registers); i++){
                    existing = 0; 
	   	    for(j = 1; j<=  dynlen(AllRegisters); j++){	   //check if register is already existing: updates or appends	
		        
	                if(AllRegisters[j] == registers[i]){ 							 
			   	dynClear(splitdata); 
	    			dynClear(splitmask); 
				fwTell1_updateVar(data[i], size[i], splitdata, splitmask);                   //fwTell1_updateVar gives back data and mask 		 
				fwTell1_writeToVar(AllRegistersData[i], AllRegistersMask1[i], splitdata, splitmask);	   //update of the data and the mask (according to the two old ones) 
				AllRegistersDataCombined[i] = AllRegistersData[i] + AllRegistersMask[i];
                                existing = 1;
                                break; 
			}
                     }
                   if(!existing){
                        fwTell1_updateVar(data[i], size[i], splitdata, splitmask);  
                        dynAppend(AllRegistersData, splitdata);
                        dynAppend(AllRegistersMask, splitmask);
                        dynAppend(AllRegistersType, type[i]);
                        dynAppend(AllRegistersSize, size[i]);
	           }
                }
	      }else{ 
 		   DebugTN("Recipe parsed without registers!?"); 
		} 
     	} //loop over recipes
   }else{ 
      DebugTN("Less than two recipes given as argument!?"); 
   } 
 
   //DebugTN(HardwRegisters_ALL);
   dynClear(registers);
   if (fwHw_createRecipeType(RecipeType,dpType1,AllRegistersType)){
      DebugN("Registers belonging to recipe type: " + RecipeType);  
      fwHw_getRecipeTypeInfo(RecipeType, dpType1, registers); 
      DebugN(registers);
   }else{                                                             //type already exists - check if order of recipes is correct
      fwHw_getRecipeTypeInfo(RecipeType, dpType1, registers); 
      for(i = 1; i <= dynlen(registers); i++){
          if(registers[i] != AllRegisters[i])
             DebugN("registers not in expected order!!!  -  " + registers[i]);   //in case it matters?
      }
   }
        

   fwHw_saveRecipe(hardware, RecipeType, RecipeName, AllRegisters, AllRegistersType, AllRegistersDataCombined); 
    
   DebugTN("=========================END=OF=THE=ASSEMBLY=============================="); 
   DebugTN("=========================================================================="); 
 
} 
 
 
 */
 


//this function is used in the recipe configurator panels
 
fwTell1_updateVar(dyn_anytype loadData,int size,dyn_anytype &Data,dyn_anytype &Mask)   
{ 
  	int i; 
	if (dynlen(loadData) == (2 * size)){ 
		for (i=1; i <= size; i++){ 
			Data[i] = loadData[i]; 
			Mask[i] = loadData[i+size]; 
		} 
 
	}else{   
		Data = loadData;       
		Mask = loadData; 
	} 
}	 
 
 
 //this function is used in the recipe configurator panels
 
fwTell1_writeToVar(dyn_anytype &RegVarData, dyn_anytype &RegVarMask, dyn_anytype NewData, dyn_anytype RegMask){ //, dyn_anytype &DataOut, dyn_anytype &MaskOut){ 
  
 int i; 
 dyn_anytype maskedRead; 
 dyn_anytype maskedWrite; 
	 
	 
 bool DEBUG = false; 
	 
	if(DEBUG){ 
		DebugTN("-----------------------------------------"); 
		DebugTN("old data : "+fwHw_convertByteToHex(RegVarData)); 
		DebugTN("old mask : "+fwHw_convertByteToHex(RegVarMask)); 
		DebugTN("-----------------------------------------"); 
		DebugTN("add data : "+fwHw_convertByteToHex(NewData)); 
		DebugTN("add mask : "+fwHw_convertByteToHex(RegMask)); 
		DebugTN("-----------------------------------------"); 
	} 
 
	if(dynlen(RegMask) != dynlen(RegVarData)) 
		DebugN("WARNING: Length of mask does not match register length!!!"); 
   
   
	for(i = 1; i <= dynlen(NewData); i++){ 
		maskedRead[i]  = RegVarData[i] & (~RegMask[i]);  //read_tmp[i] &= ~mask_tmp[i]; 
		maskedWrite[i] = NewData[i]    & RegMask[i];       //write_tmp[i] &= mask_tmp[i]; 
		RegVarData[i]  = maskedRead[i] | maskedWrite[i]; //write_tmp[i] |=  read_tmp[i];       
		RegVarMask[i]  = RegVarMask[i] | RegMask[i]; 
	} 
	 
	if(DEBUG){ 
		DebugTN("new data : "+fwHw_convertByteToHex(RegVarData)); 
		DebugTN("new mask : "+fwHw_convertByteToHex(RegVarMask)); 
		DebugTN("-----------------------------------------"); 
		DebugTN("-----------------------------------------");  
	}  
  
 //DataOut = RegVarData; 
 //MaskOut = RegVarMask; 
   
} 
 
 

// introduced to retrieve register type and size of hardware registers
 
int fwTell1_getRegInfo(string hardware, dyn_string Registers, dyn_int &Registers_Size, dyn_string &MyregType) 
{ 
	dyn_dyn_anytype list6789;  //Find out yourself list Nr 1    (specific settings) 
        dyn_string newRegisters; 
        dyn_string typeReg; 
        dyn_string dum;
       
        
        for(int i =1; i<= dynlen(Registers); i++) {
            //dynClear(dum);
            dum = strsplit(Registers[i],":");
            if(dynlen(dum) >= 2)
              Registers[i] = dum[2];
            dynAppend(newRegisters,substr(Registers[i],strlen(hardware+"."))); 
        }
  
	if(!fwHw_getSpecificSettings(hardware, newRegisters, list6789))
          return -1;
       // DebugTN("fwTell1_getRegInfo:");
       // DebugN("hardware: " + hardware);
       // DebugN("newRegisters: " + newRegisters);
       // DebugN("specific Settings: " + list6789);
        
        // it is ugly - but we can identify the type of register by the number of parameters
        for(int i = 1; i <= dynlen(Registers); i++){  
                  if(dynlen(list6789[i]) == 5){ 
			dynAppend(MyregType,FWHW_LBUS); 
			dynAppend(Registers_Size, (int) list6789[i][2] * (int) list6789[i][3]); 
		  } 
                  else if(dynlen(list6789[i]) == 9){ 
			dynAppend(MyregType,FWHW_I2C); 
			dynAppend(Registers_Size, list6789[i][4]); 
		  } 
                  else if(dynlen(list6789[i]) == 4){ 
			dynAppend(MyregType,FWHW_GBE); 
			dynAppend(Registers_Size, (int) list6789[i][2] * 4); 
		  } 
                  else if(dynlen(list6789[i]) == 5){ 
			dynAppend(MyregType,FWHW_UserSpecific); 
			dynAppend(Registers_Size, (int) list6789[i][3] * (int) list6789[i][4]); 
		  }
                  else{
                    DebugN("*** ERROR: register type could not be identified! ***"); 
                  }           
          } 
        return 0;
}  




